<?php
namespace Zeedhi\Framework\Remote\FileUploader;

use Zeedhi\Framework\DTO\Request;
use Zeedhi\Framework\Events\PreDispatch\Listener as AbstractListener;
use Zeedhi\Framework\File\Image\Image;
use Zeedhi\Framework\File\Image\Processor\Processor;

class ProcessorListener extends AbstractListener {

    protected $processor;
    protected $configuration = [];

    public function __construct(Processor $processor, array $configuration) {
        $this->processor = $processor;
        $this->configuration = $configuration;
    }

    protected function shouldCreateImages(string $route) : bool {
        return isset($this->configuration[$route]);
    }

    protected function getRequestRows(Request $request) : array {
        $rows = [];
        if ($request instanceof Request\Row) {
            $rows = [$request->getRow()];
        }

        if ($request instanceof Request\DataSet) {
            $rows = $request->getDataSet()->getRows();
        }

        return $rows;
    }

    protected function createImageFromBase64(string $b64File) : Image {
        $imageData = explode(',', $b64File);

        $image = imagecreatefromstring(base64_decode($imageData[1]));
        $type = substr($imageData[0], 5, -7);

        return new Image($image, $type);
    }

    protected function getImageWriterMethod(Image $image) : string {
        switch ($image->getMimeType()) {
            case 'image/jpeg':
                return 'imagejpeg';
            case 'image/png':
                return 'imagepng';
        }
        throw new \Exception('Invalid image type');
    }

    protected function getImageInfo(Image $image) : array {
        ob_start();

        $method = $this->getImageWriterMethod($image);
        $method($image->getImage());

        $contents = ob_get_contents();
        ob_end_clean();

        return [
            'base64' => 'data:' . $image->getMimeType() . ';base64,' . base64_encode($contents),
            'size' => strlen($contents)
        ];
    }

    protected function generateImages(array $images) : array {
        return array_map(function($image) {
            $imageObj = $this->createImageFromBase64($image['b64File']);
            $processedImage = $this->processor->apply($imageObj);

            $processedImageInfo = $this->getImageInfo($processedImage);
            return [
                'b64File' => $processedImageInfo['base64'],
                'name' => $image['name'],
                'size' => $processedImageInfo['size'],
                'type' => $processedImage->getMimeType()
            ];
        }, $images);
    }

    protected function createProcessedImage(&$row, array $configuration) {
        $imageFieldName = $configuration['originalFileField'];
        $generatedImageFieldName = $configuration['generatedFileField'];

        if (isset($row[$imageFieldName])) {
            $row[$generatedImageFieldName] = $this->generateImages($row[$imageFieldName]);
        }
    }

    protected function createProcessedImages(array &$rows, array $configuration) {
        foreach ($rows as $index => $row) {
            $this->createProcessedImage($rows[$index], $configuration);
        }
    }

    public function preDispatch(Request $request) {
        $route = $request->getRoutePath();
        if ($this->shouldCreateImages($route)) {
            $rows = $this->getRequestRows($request);

            $this->createProcessedImages($rows, $this->configuration[$route]);
        }
    }

}